clear
clc

%% Definition of the parameters of the system
N=4; % Number of full turns of the spiral
T=120; % Time to complete the spiral
A=0.5; % Radius of the outer loop of the spiral
d=0.1; % Distance between wheels
rw=2e-2; % Radius of the wheels

%% Definition of the trajectory
%  The trajectory is defined in polar coordinates; radius "r" and angle
%  "phi" as functions of time.

% This combination is my favourite so far. It requires the wheels to be
% able to turn backwards but produces a beautiful spiral with smooth
% accellerations. The scaling function is used to tune the timing. I
% obtained it by trial an error.

scaling=@(x) x.^5.*(15*x.^2-35*x+21);%-x.^4.*(20*x.^3-70*x.^2+84*x-35);
% scaling=@(x) x;
fr=@(t) (scaling(t/T)).^2*A;
fphi=@(t) 2*pi*N*sqrt(scaling(t/T));

% Other possible choices for the radius and the angle are given below. They
% are in order: you must uncomment a matching pair of fr and fphi.

% fr=@(t) A*t/T;
% fr=@(t) (-1+exp(t/T*log(2)))*A;
% fr=@(t) (t/T).^2*A;
% fr=@(t) sqrt(t/T).^2*A;
% fr=@(t) (t/T).^6*A;
% fr=@(t) (-2*(t/T).^2.*(0.5*(t/T).^2-1)).^2*A;
% fr=@(t) ((t/T).^3.*((t/T).^2-10/3*(t/T)+10/3)).^2*A;

% fphi=@(t) 2*pi*N*t/T;
% fphi=@(t) 2*pi*N*(t/T).^2;
% fphi=@(t) 2*pi*N*sqrt(t/T);
% fphi=@(t) 2*pi*N*sqrt(sqrt(t/T));
% fphi=@(t) 2*pi*N*((t/T).^(3/2));
% fphi=@(t) 2*pi*N*sqrt(-2*(t/T).^2.*(0.5*(t/T).^2-1));
% fphi=@(t) 2*pi*N*sqrt(((t/T).^3.*((t/T).^2-10/3*(t/T)+10/3)));

% The code below draws an L shape just to demonstrate the limitations of
% the rudimentary path-planning presented here: some paths may require the
% vehicle's angle to change instantaneously, which requires the wheels to
% suffer an impulse-type change

% fx=@(x) (x>=0.5).*(x-0.5)*2*A;
% fy=@(x) (x<0.5).*(x)*2*A+(x>=0.5)*A;
% scaling=@(x) (x/T).*((x/T).^2-1.5*(x/T)+3/4)*4;
% fr=@(x) sqrt(fx(scaling(x)).^2+fy(scaling(x)).^2);
% fphi=@(x) atan2(fy(scaling(x)),fx(scaling(x)));

%% Generation of simulation data
%  The code below simulates the motion of the vehicle and generates arrays 
%  of time series of different relevant variables that are plotted.

Nsteps=1000; 
dt=T/(Nsteps-1);
t=0:dt:T; % This is the clock of the simulation

r=fr(t);
phi=fphi(t);
sphi=sin(phi);
cphi=cos(phi);

% (x,y) are the coordinates of the centre of the vehicle.
x=r.*cos(phi);
y=r.*sin(phi);

% This is how we calculate the speed of the vehicle.
dx=diff(x)/dt; 
dx=[dx,dx(end)];
dy=diff(y)/dt;
dy=[dy,dy(end)];
v=sqrt(dx.^2+dy.^2); 

% We also want to know how fast the radius and angle of the spiral change.
dr=diff(r)/dt;
dr=[dr,dr(end)];
dphi=diff(phi)/dt;
dphi=[dphi,dphi(end)];

%% The big picture
% The code below opens a big plot window to show a number of relevant variables.

% (x,y) position against time [shows the actual spiral]
figure('Name','Many plots','Position',[100 100, 1200, 800]);
subplot(3,4,1);
plot(x,y);
xlabel('X position [m]');
ylabel('Y position [m]');
grid on;
axis([-A,A,-A,A]);

% speed against time
subplot(3,4,2);
plot(t,v);
ylabel('Speed [m/s]');
xlabel('Time [s]');
grid on;

% radius and rate of change of radius against time
subplot(3,4,4);
[hAx,hLine1,hLine2]=plotyy(t,r,t,dr);
ylabel(hAx(1),'Radius [m]');
ylabel(hAx(2),'dRadius/dt [m/s]');
clear hAx hLine1 hLine2
xlabel('Time [s]');
grid on;

% angle and rate of change of the angle against time
subplot(3,4,5);
[hAx,hLine1,hLine2]=plotyy(t,phi*180/pi,t,dphi*180/pi);
grid on;
ylabel(hAx(1),'\phi [deg]');
ylabel(hAx(2),'d\phi/dt [deg/s]');
clear hAx hLine1 hLine2
xlabel('Time [s]');

% x coordinate against time (not very interesting)
subplot(3,4,7);
plot(t,x);
ylabel('X position [m]');
xlabel('Time [s]');
grid on;

% y coordinate against time (not very interesting)
subplot(3,4,8);
plot(t,y);
grid on;
ylabel('Y position [m]');
xlabel('Time [s]');

% The code above only tells us how the centre of the vehicle will move, but
% doesn't say much about how the wheels will move nor how the chassis will
% rotate. The first thing we must find out is how the orientation of
% the chassis will change over time:

stheta=(-dr.*sin(phi)-r.*dphi.*cos(phi))*(-1)./sqrt(dr.^2+r.^2.*dphi.^2);
ctheta=(-dr.*cos(phi)+r.*dphi.*sin(phi))*(-1)./sqrt(dr.^2+r.^2.*dphi.^2);
% temp=sqrt(stheta.^2+ctheta.^2);
% stheta=stheta./temp;
% ctheta=ctheta./temp;
% clear temp;
theta=atan2(stheta,ctheta);
temp=diff(theta);
temp=temp+(temp<(-2*pi*0.9))*2*pi;
theta=[theta(1),theta(1)+cumsum(temp)];
clear temp;

% Next we calculate the angular acceleration of the chassis:
dtheta=diff(theta)/dt;
dtheta=[dtheta,dtheta(end)];

% Here we calculate the distance the centre of the vehicle has traveled
len=cumsum(v)*dt;

% Distance traveled against time
subplot(3,4,9);
plot(t,len);
grid on;
xlabel('Time [s]');
ylabel('Distance [m]');

% Here we calculate how the x and y position of the centre of the vehicle.
% This should match the (x,y) data points we defined earlier.
dxc=v.*ctheta;
dyc=v.*stheta;
xc=cumsum(v.*ctheta)*dt;
yc=cumsum(v.*stheta)*dt;
subplot(3,4,1);
hold on;
plot(xc,yc);

% Knowing the speed of the centre of the chassis and how the chassis
% rotates, we can obtain the speed of the centre of the two weels (right
% and left):
VR=v+d*dtheta/2;
VL=v-d*dtheta/2;

% Speed of the centre of the wheels against time
subplot(3,4,3);
plot(t,VR);
hold on;
plot(t,VL);
plot(t,v);
grid on;
xlabel('Time [s]');
ylabel('Speed [m/s]');
legend('R','L','centre');

% Knowing the speed of the centre of the wheels and their radius, we can
% obtain the angular speed of the wheels and how their angle of rotation
% varies over time (we assume they don't slide):
wr=VR/rw;
wl=VL/rw;
thetar=cumsum(wr)*dt;
thetal=cumsum(wl)*dt;

% Angle of roration of each wheel against time
subplot(3,4,10);
plot(t,thetar*180/pi);
hold on;
plot(t,thetal*180/pi);
xlabel('Time [s]');
ylabel('Angle [deg]');
grid on;
legend('R','L');

% We can calculate the angular accelleration of each wheel. If this was too
% high, the wheels would be likely to slide:
dVR=diff(VR);
dVR=[dVR,dVR(end)];
dVL=diff(VL);
dVL=[dVL,dVL(end)];

% Acceleration of the centre of each wheel against time:
subplot(3,4,6);
plot(t,dVR);
hold on;
plot(t,dVL);
grid on;
xlabel('Time [s]');
ylabel('Accel [m/s^2]');
legend('R','L');

%% Animation
%  Here we produce a nice 3D animation showing how the vehicle moves. Some
%  of the 3D plot functions used here are not supported in GNU Octave,
%  therefore the following code will only work in Matlab. The trajectory of 
%  the centre of the vehicle marked on the floor. The trajectory of one
%  point of each wheel is also marked as a trace in the air.

figure('Name','3D animation','Position',[10,10,1000,500*0.8]);
[x1,x2,x3,y1,y2,y3]=calc_coords(xc(1),yc(1),theta(1),0.1);
p=fill3([x1,x2,x3],[y1,y2,y3],[rw,rw,rw],[0.9;0;0]);
mr=coord_wheel(x2,y2,thetar(1),theta(1),rw);
ml=coord_wheel(x3,y3,thetal(1),theta(1),rw);
hold on
pr=draw_wheel(mr);
pl=draw_wheel(ml);
xlabel('x [m]');
ylabel('y [m]');
zlabel('z [m]');
axis([-A*1.1,A*1.1,-A*1.1,A*1.1,0,A*1.1*0.8]);
set(gca,'XTick',sort([-A*1.1:(A/5):A*1.1,0]));
set(gca,'YTick',sort([-A*1.1:(A/5):A*1.1,0]));
grid on
view([-2,-2,2]);
m=animatedline();
mmr=animatedline('Color',[0.8,0.6,0.6]);
mml=animatedline('Color',[0.6,0.8,0.6]);
%m=plot3(xc,yc,zeros(1,length(xc)));
tic
for n=1:length(t)
%     delete(m);
%     m=plot3(xc(1:n),yc(1:n),zeros(1,n));
    m.addpoints(xc(n),yc(n),0);
    [x1,x2,x3,y1,y2,y3]=calc_coords(xc(n),yc(n),theta(n),0.1);  
    p.Vertices(1,1)=x1;
    p.Vertices(1,2)=y1;
    p.Vertices(2,1)=x2;
    p.Vertices(2,2)=y2;
    p.Vertices(3,1)=x3;
    p.Vertices(3,2)=y3;
    mr=coord_wheel(x2,y2,thetar(n),theta(n),rw);
    ml=coord_wheel(x3,y3,thetal(n),theta(n),rw);
    mmr.addpoints(mr(1,2),mr(2,2),mr(3,2));
    mml.addpoints(ml(1,2),ml(2,2),ml(3,2));
    update_wheel(mr,pr);
    update_wheel(ml,pl);
    drawnow

%   The code below can be used to export the frames as bitmaps that can be
%   combined into an gif animation or a video.

%     if (mod(n,10)==1)
%         print(['./Frames/f' num2str(floor(n/10),'%04d') '.png'],'-dpng');
%     end
end
toc

